!(c)2001 copyright Tim Hatamian Mathematicus Laboratories
!Last Revision Dec-18-2001 (tested on CVF only)


!Autoarrays are self-adjusting arrays who autoamtically reallocate themselves as the need arises
!without explicit intervnetiuon from the user. Autoarrays are an easy alternative to linked lists.
!This version implements only REAL(4) types.

! they are created with two(or three) paramerters:
!			initial size, the futures increment, and an optional startIndex (DEFAULT=1)
! the array is first allocated with the initial size, thereupon whenever an index of the array
! is written-to which is out of its current bounds the array automatically resizes itself
! to the the assigned index plus the increment size.
! For example suppose array A is currently sized 5 with an dSize(increment) of 2
! When we attempt to assign a value to index 6, array-A is first extended to size 6+2=8 then the 
! assignment proceeds as usual.
! If the curent size is 8, try to access element 10 then the size becomes 12.
! if the size is 12 assining element 11 does NOT change the size

! It is advised that the user make the best estimate of the size intially so as to MINIMIZE the number of
! resizings if the final application is time intensive.

! If you find this useful don't feel shy about sending me a thank you note on the back of a $5 bill, or via Paypal.
! My Paypal email address is tsh@mathematicuslabs.com
! If you do, I will send you the Module source code. By then I may have implemented other types as well.
! Please forward any bugs or comments to the same address.
  

!dec$ if(.false.)	!this is a CVF command (if using another compiler comment out this section some other way)
! list of available methods
!------------------------------
!constructor
! if calling with "startIndex" but omitting "lastErr" call must be with keyword, e.g:
! CALL new_AutoArray(AAX,10,5,startIndex=-2) 
! Watch out for size==0 or dSize==0; these are the most common reasons for trouble.
	elemental subroutine new_AutoArray(AA, size, dSize [,lastErr] [,lowBound=LowBoundValue])
		type(Real4AutoArray),intent(inout):: 	AA
		integer,intent(in)::					Size,dSize
		integer,intent(in),optional::			lowBound

!destrcutor (note: on arrays of AutoArrays you MUST also deallocate the array manually AFTER this call)
	elemental subroutine destroy_AutoArray(AA, [lastErr])			
		type(Real4AutoArray),intent(inout):: 	AA

!Accessor; can also use "AA>index"  
!Note: AA>i=2 will NOT work: The accessor ">" is read-only (see AA_assignVal)
!Returns ZERO if index out of current defined range
!Returns NaN if AA not yet constructed
	elemental real(4) function AA(AA, index )						
		type(Real4AutoArray),intent(in):: 		AA
		integer,intent(in)::					index

!alternate Accessor
!Returns ZERO if index out of current defined range
!Returns NaN if AA not yet constructed
	elemental real(4) function AA_get(AA, index )					
		type(Real4AutoArray),intent(in):: 		AA
		integer,intent(in)::					index

!current size of the array
	elemental integer function AA_getCurrSize(AA, [lastErr])		
		type(Real4AutoArray)::					AA

!current Increment size of the array
	elemental integer function AA_getIncrSize(x [,lastErr])
		type(Real4AutoArray)::					AA

! returns the highest index ever assigned a value
	elemental integer function AA_getHighIndex(AA, [lastErr])
		type(Real4AutoArray),intent(in)::		AA

! the number of bins assigned since creation
! returns NaN on error
	elemental integer function AA_binsAssigned(AA)
		type(Real4AutoArray),intent(in)::		AA

!Reset the AutoArray to its setting as of its creation. All elements=0.
	elemental subroutine AA_reset(AA, [lastErr])
		type(Real4AutoArray),intent(inout):: AA

! change the value of increment size
	elemental subroutine AA_resetIncrSize(AA, dSize, [lastErr])
		type(Real4AutoArray),intent(inout):: 	AA
		integer,intent(in)::					dSize

! resize to n+dSize if n>AA.currSize  
! n<= currSize results in no action
	elemental subroutine AA_resetCurrSize(AA, newSize, [lastErr])
		type(Real4AutoArray),intent(inout):: 	AA
		integer,intent(in)::					newSize

! truncates to n if n<currSize
	elemental subroutine AA_cutCurrSize(x,  newSize, [lastErr])
		type(Real4AutoArray),intent(inout)::	AA
		integer,intent(in)::					newSize

!scalar assignment to an individual element
!Note: AA>i=2 will NOT work. The accessor ">" is read-only
	elemental subroutine AA_assignVal(AAtarget, index, SourceScalar)
		type(Real4AutoArray),intent(inout)::	AAtarget
		integer,intent(in)::					index
		integer(4,2,1), real(4,8),intent(in)::	SourceScalar

!Alternate scalar assignment to an individual element(strict typed) for faster access
	elemental subroutine AA_assignRealVal(AAtarget, index, SourceScalar)
		type(Real4AutoArray),intent(inout)::	AAtarget
		integer,intent(in)::					index
		real(4),intent(in)::					SourceScalar

!---------------computations-------------
!Sort: 
! Sorts the AutoArray starting with the lowBound, but only up to the highestIndex assigned.
!Notes:
! 1. Any unassigned elements (while still =0) remain at their original positions.
! 2. Returns NaN if AA is undefined (typically because new_AutoArray was called w/ size==0 or dSize==0)
! 3. This method is not elemental (Cannot pass arrays of AA's to it)
! 4. If the optional argument (AAsort) is elliminated, the input (AAin) is overwritten with the sorted values.
	real(4) function AA_sort(AAin,[AAsort])

!Sum: returns NaN if AA-undefined or other invalid
	elemental real(4) function AA_sum(AA)

!Average: returns NaN if x-undefined or other invalid
	elemental real(4) function AA_ave(AA)

!The nth-Moment: returns NaN for n<0 or other invalid
	elemental real(4) function AA_moment(AA,n)
		integer,intent(in):: n

!The nth-Cumulant: returns NaN for n<0 or other invalid
	elemental real(4) function AA_cumulant(AA,n)
		integer,intent(in):: n


!----------------------------------------
! elemental "=" array assignment to AA starting at index-1
!	AAtarget = sourceArray
!	sourceArray = AAtarget	(watch for truncation, overflow, loss of precision into lower types)
!	AAtarget = AAtarget
		type(Real4AutoArray),intent(inout):: AAtarget
		integer(4,2,1) real(4,8),intent(in):: sourceArray(:)
 
!-----------Common items--------------------
! "lastErr" is either optional or not present, always has this interface
		character(32),intent(out),optional::	lastErr

!NaN: for com[iler which do not have a function check for a NaN you can use this NaN in the module to check
! to see if a Nan has been returned. This has only been tested on CVF 6.6.
! for exmaple in CVF you'd use: if(ISNAN(AA>i))
! which you can replace with	if((AA>i) == AA_rNaN)
		real,parameter::	AA_rNaN


!=================================================================================================================
! Type Distribution
! While only the upper end of an AA is flexible, its lowerbound is fixed once the AA is created.
! if both ends need to be autoamtically extended then you must use the "Distribution" type.
! The "Distribution" type is comprised simply of TWO AutoArrays.
! Both of these AA's are public and via the names:
! "Dist_name".posSide   and  "Dist_name".negSide 
! thus all AA functions can be applied to each, However, you should not destroy these individually.
! "Dist_name".posSide   has a low bond of 0 while  "Dist_name".negSide has a lowBound of 1 (which emulates -1).
! NOTE: NO assignement (=) operator is available for Distributions but there is a Dist_copy method.


! constructs a Distribution, on subsequent calls increments the size in positive or negative direction
! Lower bounds are fixed at 0 for the positive side and (-)1 on the neg side.
! Watch out for size==0 or dSize==0; these are the most common reasons for trouble.
	elemental subroutine new_Distribution(Distname,size,dSize [,lastErr])
		type(Distribution),intent(inout):: Distname
		integer,intent(in):: size,dSize
		character(32),intent(out),optional:: lastErr

!Use this instead of the destroy_AutoArray to deallocate Distributions
	elemental subroutine destroy_Distribution(Dist [,lastErr])
		type(Distribution),intent(inout):: Dist
		character(32),intent(out),optional:: lastErr

!Accessor; can also use "Dist_name">index  
!Note: Dist>i=2 will NOT work The accessor ">" is read-only (see Dist_assignVal)
!Returns ZERO if index out of current defined range
!Returns NaN if Distname not yet constructed
	elemental real(4) function Dist(Distname, index )						
		type(Distribution),intent(in):: 		Distname
		integer,intent(in)::					index

!alternate Accessor
!Returns ZERO if index out of current defined range
!Returns NaN if Distname not yet constructed
	elemental real(4) function Dist_get(Distname, index )				
		type(Distribution),intent(in):: 		Distname
		integer,intent(in)::					index

!scalar assignment to an individual element
!Note: Distname>i=2 will NOT work. The accessor ">" is read-only
	elemental subroutine Dist_assignVal(Disttarget, index, SourceScalar)
		type(Distribution),intent(inout)::	Disttarget
		integer,intent(in)::					index
		integer(4,2,1), real(4,8),intent(in)::	SourceScalar

!Alternate scalar assignment to an individual element(strict typed) for faster access
	elemental subroutine Dist_assignRealVal(Disttarget, index, SourceScalar)
		type(Distribution),intent(inout)::	Disttarget
		integer,intent(in)::					index
		real(4),intent(in)::					SourceScalar

!returns highest negative index assigned since creation(always a negative integer)
	elemental integer(4) function Dist_negLim(Dist)
		type(Distribution),intent(inout)::	Dist

!returns highest positive index assigned since creation(a;lways a positive integer)
	elemental integer(4) function Dist_posLim(Dist)
		type(Distribution),intent(inout)::	Dist

!copies one Dist to another Both Dist's must already have been created
	elemental subroutine Dist_copy(targDist,srcDist)
		type(Distribution),intent(in):: srcDist
		type(Distribution),intent(inout):: targDist

!---------------computations-------------
! total number of bins assigned since creation.
	elemental integer(4) function Dist_binsAssigned(x)
		implicit none
		type(Distribution),intent(in):: x
!sum
	elemental real(4) function Dist_sum(x)
		type(Distribution),intent(in):: x

!average
	elemental real(4) function Dist_ave(x)
		type(Distribution),intent(in):: x

!nth-moment
!returns Nan for n<0 or other invalid
	elemental real(4) function Dist_moment(x,n)
		type(Distribution),intent(in):: x
		integer,intent(in):: n

!nth-Cumulant
!returns Nan for n<0 or other invalid
	elemental real(4) function Dist_cumulant(x,n)
		implicit none
		type(Distribution),intent(in):: x
		integer,intent(in):: n




!dec$ endif		!this is a CVF command (if using another compiler, comment out the above section some other way)
!================================================================================================================


Program DemoAutoArrays
	call demo_AutoArray
	call demo_MoreAutoArray
	call demo_ArrayofAutoArrays1
	call demo_ArrayofAutoArrays2
	call demo_distributions
end Program DemoAutoArrays


!============================================================================
! A single autoarray is demonstrated.
subroutine demo_AutoArray
	use AutoArrays

	type(Real4AutoArray) x
	character(32) lastErr					!container for optional error string

	write(*,*)'----------demo a single AutoArray-------------------'
	call new_AutoArray(x,5,2,lastErr,lowBound=0)		!create a new autoarray of type real(4)  (Error checking is optional)

	if(lastErr/='NONE')then					!If no errors occur lastErr is set to 'NONE' (case sensitive)
		type *, lastErr, 'During AA creation'
		pause								!most routines have optinoal error checking.
		stop								!routines with frequent access do not
	endif

	!use the method AA_getCurrSize to get the current dimension on the autoarray
	write(*,*)'AutoArray Current size=',AA_getCurrSize(x)
	write(*,*)'Highest index assigned=',AA_getHighIndex(x)
	write(*,*)(x>i,i=1,AA_getHighIndex(x)) !use the method AA_getHighIndex to get the highest index of the autoarray actually assigned a value.
	write(*,*)(AA_get(x,i),i=1,AA_getHighIndex(x)) !use the method AA_getHighIndex to get the highest index of the autoarray actually assigned a value.
	!note that at this points the highest index accessed is 0 therefore no output is expected from these lines

	!here we set all its components to 1.00000
	x=1								!mixed mode assignments are allowed
										
	call AA_assignVal(x,7,2.)		!the individual components of autoarray are assigned by the AA_assignVal
	write(*,*)AA_getCurrSize(x)
	write(*,*)(x>i,i=AA_getLowBound(x),AA_getHighIndex(x))

!Rules of precedence for the accessor operator ">"
! Like other defined operators it has the LOWEST percedence of all oprators.
! judicious use of parentheses is therefore necessary.
	write(*,*)
	write(*,*)(x>(i+1),i=5,7)
	write(*,*)
	write(*,*)(x>i+1,i=5,7)		!this is the same as the above: + takes precedence over >
	write(*,*)
	write(*,*)((x>i)+1,i=5,7)	!this one is different
	write(*,*)

	call AA_assignVal(x,8,3.)		!This call result in no size chg as 8<currSize=9, only the assignment is done
									!the index (2nd arg of assignVal) is always of type int4 
									!but the 3rd arg can be real(4,8) or int(4,2,1) 

	
	call AA_resetIncrSize(x,1,lastErr)		! change the value of increment size
	call AA_assignVal(x,10,4.)		!set element 10 and bump current size to 11
	write(*,*)AA_getCurrSize(x)
	write(*,*)(x>i,i=AA_getLowBound(x),AA_getHighIndex(x))	!the range is i=0,10
	write(*,*)

	!allocate more or less memory on the fly:
	! AA_resetCurrSize resize to n+dSize if n>currSize  
	call AA_cutCurrSize(x,4,lastErr)		!cut some elements, current size=4+1=5 (n>= currSize results in no action)
	write(*,*)AA_getCurrSize(x)
	call AA_resetCurrSize(x,5,lastErr)		!more elements, current size=5+1=6		(n<= currSize results in no action)
	write(*,*)
	write(*,*)AA_getCurrSize(x)
	write(*,*)(AA_get(x,i),i=AA_getLowBound(x),AA_getHighIndex(x))	!some previous elements were lost when reset to smaller sizes

	call destroy_AutoArray(x,lastErr)
	if(lastErr /= 'NONE' )then
		type *, lastErr, 'during AA destroy'
		pause 
		stop
	endif


	pause
end subroutine demo_AutoArray

!============================================================================
subroutine demo_MoreAutoArray
	use AutoArrays

	type(Real4AutoArray) AAx,AAy
	real,allocatable::		y(:)	!regular primitive type arrays, can also be fixed length.
	integer,allocatable::	n(:)	!regular primitive type arrays, can also be fixed length.
	character(32) lastErr					!container for optional error string


	write(*,*)'----------demo a single AutoArray: copy ops between regular and autoarrays -------------------'
	call new_AutoArray(AAx,5,2,lastErr)		!create a new autoarray of type real(4)
	call new_AutoArray(AAy,5,2,lastErr)		!create a new autoarray of type real(4)
	allocate(n(5),y(5))
	AAy=2.
	AAx=AAy
	write(*,*)(AA_get(AAx,i),i=1,AA_getHighIndex(AAx))
	n=3
	AAx=n					!mixed mode assignment to AutoArray is allowed
	write(*,*)
	write(*,*)(AA_get(AAx,i),i=1,AA_getHighIndex(AAx))

	y=AAx					!mixed mode assignment to AutoArray is allowed
	write(*,*)
	write(*,*)(y(i),i=1,size(y))

	AAx=int2(1)				!while Autoarryas are of type real(4) they can be assigned from or to int4, int2, int1, real8
	n=AAx					!be careful about overflows in going to smaller types like Int2, int1 
	write(*,*)
	write(*,*)(n(i),i=1,size(n))

	y=[5.,2.,3.,1.,0.]
	do i=1,2
		AAx=real(i)*y
		if(i==1)call AA_assignVal(AAx,6,4.)
		if(isnan(AA_sort(AAx)))stop 'Invalid AA passed to Sort'
		write(*,*)
		write(*,*)(AA_get(AAx,j),j=1,AA_getHighIndex(AAx))
		call AA_reset(AAx)		!IMPORTANT: without this reset, the values from the previous iter.
								! may be included in this iter. since highestIndex is not reset.
	enddo

	call destroy_AutoArray(AAx,lastErr)
	call destroy_AutoArray(AAy)			!error checking is optional

	PAUSE

end subroutine demo_MoreAutoArray

!============================================================================
subroutine demo_ArrayofAutoArrays1
	use AutoArrays
	type (Real4AutoArray),allocatable,volatile:: AAarray(:)
	integer,allocatable:: narr(:)
	character(32)	lastErr
	character(32),ALLOCATABLE::	lastErrs(:)

	allocate(AAarray(3))
	allocate(narr(3),lastErrs(3))
	write(*,*)'----------demo Array of AutoArrays1------------'
	do i=1,3
		write(*,*)'----------Element:',i,'----------------'
		call new_AutoArray(AAarray(i),3,2,lastErr)
		AAarray(i)=i
		call AA_assignVal(AAarray(i),5,2.)
		write(*,*)AA_getCurrSize(AAarray(i))
		write(*,*)(AA_get(AAarray(i),j),j=1,AA_getHighIndex(AAarray(i)))
		call AA_assignVal(AAarray(i),6,real(i))

		write(*,*)AA_getCurrSize(AAarray(i))
		write(*,*)(AA_get(AAarray(i),j),j=1,AA_getHighIndex(AAarray(i)))
	enddo
	narr=AA_getCurrSize(AAarray)
	write(*,*)narr
	call destroy_AutoArray(AAarray(1),lastErr)
	write(*,*)'------------------element 1 after deallocation------------'
	write(*,*)(AA_get(AAarray(1),j),j=1,AA_getHighIndex(AAarray(1)))
	write(*,*)'----------------------element 2----------------'
	write(*,*)(AA_get(AAarray(2),j),j=1,AA_getHighIndex(AAarray(2)))
	write(*,*)'----------------------element 3----------------'
	write(*,*)(AA_get(AAarray(3),j),j=1,AA_getHighIndex(AAarray(3)))
	call destroy_AutoArray(AAarray,lastErrs)		!note that when an Array of AAs is deallocated (or otherwise processed)
											!each element will have its own error code, hence the array lastErrs.
!An alternative would be a loop over  
!	call destroy_AutoArray(AAarray(i),lastErr)
! or w/o error checking
!	call destroy_AutoArray(AAarray(i))
!or
!	call destroy_AutoArray(AAarray)

	deallocate(AAarray)							!Finally the user must deallocate his Array. 
											!The revers order will result in memory leaks

	pause

end subroutine demo_ArrayofAutoArrays1
!===========================================================================================================
!
subroutine demo_ArrayofAutoArrays2
	use AutoArrays
	type (Real4AutoArray),allocatable,volatile:: AAarray(:)
	type (Real4AutoArray):: AAarray2(3)
	integer,allocatable:: narr(:)
	real(8)		aDPArray(3)

	allocate(AAarray(3))
	allocate(narr(3))
	write(*,*)'----------demo Array of AutoArrays(elemental properties)------------'

	call new_AutoArray(AAarray,3,2)		!allocate all elements at once to the same size.
	AAarray=[1.0,2.0]							!assign a scalar to each element result is arrays of 1's, 2's
												!the third elements remains all zeros
	do i=1,3
		write(*,*)(AA_get(AAarray(i),j),j=1,AA_getHighIndex(AAarray(i)))
	enddo

	narr=[4,5,6]								!another way of assigning scalar values to an AA
	AAarray=narr
	write(*,*)
	do i=1,3
		write(*,*)(AA_get(AAarray(i),j),j=1,AA_getHighIndex(AAarray(i)))
	enddo

	call AA_assignVal(AAarray,2,5.)			!the 2nd element of each AA is set to 5.
	write(*,*)
	do i=1,3
		write(*,*)(AA_get(AAarray(i),j),j=1,AA_getHighIndex(AAarray(i)))
	enddo

	narr=[7.,8.,9.]
	call AA_assignVal(AAarray,2,narr)		!the 2nd element of each AA is set to an element of NARR
	write(*,*)
	do i=1,3
		write(*,*)(AA_get(AAarray(i),j),j=1,AA_getHighIndex(AAarray(i)))
	enddo

	write(*,*)
	do i=1,3
		aDPArray=AAarray(i)		!Go back and forth between all F90 types
		write(*,*)(aDPArray(j),j=1,3)
	enddo

	call new_AutoArray(AAarray2,4,2)		!allocate all elements at once to the same size.
	AAarray2=AAarray		!or between arrays of AA's
	write(*,*)
	do i=1,3
		write(*,*)(AA_get(AAarray2(i),j),j=1,AA_getHighIndex(AAarray2(i)))
	enddo



	call destroy_AutoArray(AAarray)		!note that when an Array of AAs is deallocated (or otherwise processed)
	deallocate(AAarray)							!Finally the user must deallocate his Array. 

	pause

end subroutine demo_ArrayofAutoArrays2
!===========================================================================================================

subroutine demo_distributions
	use AutoArrays
	type (Distribution),allocatable,volatile:: DistArray(:)
	type (Distribution):: DistArray2(3), aDist
	character(32)	lastErr
	integer i,k
	real x
	
	write(*,*)'-------------Distribution type demo---------------'
	
	call new_Distribution(aDist,5,2,lastErr)
	if(lastErr /='NONE')then
		type *, lastErr, 'During Distribution creation'
		pause								
		stop								
	endif

! A distribution is like a two headed AutoArray
! Inf act each component of the Distribution is an AutoArray and can be assigned as such
! as AA's, the negSide always has a lowBound of 1 and posSide Always has 0.
	aDist.negSide=[0, 2, 1, 4, 0]
	aDist.posSide=[1.0, 2.0, 3.5]
	write(*,*)'A sample Distribution with both ends adjustable'
	do i=-AA_getHighIndex(aDist.negSide) , AA_getHighIndex(aDist.posSide)	!note the negative sign
!	do i=Dist_negLim(aDist.negSide) , Dist_posLim(aDist.posSide)		!This is identical to the above line
		write(*,*)i,(aDist>i)
	enddo

	aDist.negSide=0	!reset
	aDist.posSide=0.
	write(*,'(/a)')'A random distribution'
	do i=1,100
		call random_number(x)
		k=5*(2.*x-1.)
		call Dist_assignVal(aDist,k,(aDist>k)+1)	!increment the counter for each occurence at bin-k
	enddo
	write(*,*)(aDist>i,i=Dist_negLim(aDist),Dist_posLim(aDist))				!3 versions of accessor
!	write(*,*)(Dist_get(aDist,i),i=Dist_negLim(aDist),Dist_posLim(aDist))
!	write(*,*)(Dist(aDist,i),i=Dist_negLim(aDist),Dist_posLim(aDist))
	write(*,*)
	write(*,*)'mean    =',Dist_moment(aDist,1)			!the second moment
	write(*,*)'variance=',Dist_moment(aDist,2)			!the second moment

	write(*,'(/a)')'The sum of a Dist= sum of its PosSide + sum of its negSide'
	write(*,*)Dist_sum(aDist),'=',AA_sum(aDist.negSide),'+',AA_sum(aDist.posSide)
	pause

	allocate( DistArray(3))
	call new_Distribution(DistArray,5,2)
	call Dist_copy(DistArray(1),aDist)
	write(*,*)'A copy of the above Dist'
	write(*,*)(DistArray(1)>i,i=Dist_negLim(DistArray(1)),Dist_posLim(DistArray(1)))


	do i=Dist_negLim(aDist),Dist_posLim(aDist)
		call Dist_assignVal(DistArray(2),i,2.*(aDist>i))
		call Dist_assignVal(DistArray(3),i,3.*(aDist>i))
	enddo
	write(*,*)(DistArray(2)>i,i=Dist_negLim(DistArray(2)),Dist_posLim(DistArray(2)))
	write(*,*)(DistArray(3)>i,i=Dist_negLim(DistArray(3)),Dist_posLim(DistArray(3)))

	write(*,*)'Another copy Dist from Dist-Array to array (elemental)'
	call new_Distribution(DistArray2,5,2)				!create a fixed array of the Dist's
	call Dist_copy(DistArray2,DistArray)				!copy will be elemental or...
	do k=1,3
!		call Dist_copy(DistArray2(k),DistArray(k))		!as in here  you can state it explicitly in a loop 
		write(*,*)(DistArray2(k)>i,i=Dist_negLim(DistArray2(k)),Dist_posLim(DistArray2(k)))
	enddo

!clean up
	call destroy_Distribution(aDist)
	call destroy_Distribution(DistArray2)
	call destroy_Distribution(DistArray)	!note 2 step deallocation for allocatable array of Dist's
	deallocate(DistArray)

	pause

end subroutine demo_distributions
!===========================================================================================================
